import pandas as pd
from hedron import cluster_functions as cl
from .maps import plot_cluster, plot_super_cluster, plot_heat_map


class Cluster(pd.DataFrame):
    """Holds a pandas DataFrame with coordinate data"""
    def __init__(self,
                 df,
                 lat_header,
                 lon_header,
                 date_time_header,
                 id_header,
                 colors=None):
        pd.DataFrame.__init__(self, df)
        self.lat_header = lat_header
        self.lon_header = lon_header
        self.id_header = id_header
        self.date_time_header = date_time_header
        self.day_header = 'day'
        self.colors = colors
        if len(df) == 0: return

        # Try to convert columns to correct data types
        self[lat_header] = self[lat_header].astype(float)
        self[lon_header] = self[lon_header].astype(float)
        self[date_time_header] = pd.to_datetime(df[date_time_header])
        self[id_header] = self[id_header].astype(str)
        # Add day column
        df[self.day_header] = df[date_time_header].dt.date

    @property
    def lats(self):
        return self[self.lat_header]

    @property
    def lons(self):
        return self[self.lon_header]

    @property
    def ids(self):
        return self[self.id_header]

    @property
    def dates(self):
        return self[self.date_time_header]

    @property
    def days(self):
        return self[self.day_header]

    def plot(self, size=(800, 500)):
        return plot_cluster(self, size=size)

    def make_clusters(self, digits):
        if len(self)==0: return SuperCluster(dict())
        return convert_dict_to_super(self,
                                     cl.cluster_coords(self,
                                                       self.lat_header,
                                                       self.lon_header,
                                                       digits),
                                    colors=self.colors)

    def colocation_clusters(self, digits):
        if len(self)==0: return SuperCluster(dict())
        return convert_dict_to_super(self,
                                     cl.colocation_cluster_coords(self,
                                                                  self.lat_header,
                                                                  self.lon_header,
                                                                  self.id_header,
                                                                  digits),
                                    colors=self.colors)

    def day_colocation_cluster(self):
        if len(self)==0: return self
        return Cluster(cl.day_colocations(self,
                                          self.day_header,
                                          self.id_header),
                       self.lat_header,
                       self.lon_header,
                       self.date_time_header,
                       self.id_header,
                       colors=self.colors)

    def day_colocation_clusters(self):
        if len(self)==0: return SuperCluster(dict())
        return convert_dict_to_super(self,
                                     cl.day_colocations(self,
                                                        self.day_header,
                                                        self.id_header,
                                                        merge=False),
                                     colors=self.colors)


class SuperCluster(dict):
    """Holds multiple Cluster Objects"""

    def __init__(self, iterable, colors=None):
        dict.__init__(self, iterable)
        self.colors = colors

    def plot(self):
        return plot_super_cluster(self)

    def plot_heat(self, p):
        return plot_heat_map(self, p)

    def clusters(self):
        return self.values()

    def names(self):
        return self.keys()

    def colocation_clusters(self):
        if len(self)==0: return self
        return SuperCluster({key:cluster for key, cluster in self.items() if len(cluster[cluster.id_header].unique())>1})

    def merge(self):
        # TODO: Combine each cluster into one cluster
        pass

    def day_colocation_clusters(self):
        if len(self)==0: return self
        day_clusters = dict()
        for key, cluster in self.items():
            c = cluster.day_colocation_cluster()
            if len(c) > 0:
                day_clusters[key] = c
        return SuperCluster(day_clusters)

    @property
    def ids(self):
        out = []
        for c in self.values():
            out.extend(list(c.ids))
        return out

    @property
    def lats(self):
        out = []
        for c in self.values():
            out.extend(list(c.lats))
        return out

    @property
    def lons(self):
        out = []
        for c in self.values():
            out.extend(list(c.lons))
        return out

    # TODO: to_xlsx method, stores each cluster in a tab. saves in xlsx file

def convert_dict_to_super(cluster, d, colors=None):
    if len(d)==0: return SuperCluster(dict())
    return SuperCluster({key:Cluster(df,
                                     cluster.lat_header,
                                     cluster.lon_header,
                                     cluster.date_time_header,
                                     cluster.id_header) for key, df in d.items()})


import pandas as pd
from pygeodesy import geohash


geohash_data = {}

def calculate_geohashes(lats, lons, precision):
    hashes = []
    for lat, lon in zip(lats, lons):
        if (lat, lon, precision) in geohash_data:
            h = geohash_data[(lat, lon, precision)]
        else:
            h = geohash.encode(lat, lon, precision)
            geohash_data[(lat, lon, precision)] = h
        hashes.append(h)
    return hashes


def day_colocations_clusters(clusters, day_header, id_header):
    """Check each cluster for ids on same day"""
    out = dict()
    for key,df in clusters.items():
        day_co = day_colocations(df, day_header, id_header)
        if len(day_co) > 0:
            out[key] = day_co
    return out


def day_colocations(cluster, day_header, id_header, merge=True):
    cluster = cluster.copy()
    day_clusters = cluster.groupby(day_header)
    colocated = {key:df for key, df in day_clusters if len(df[id_header].unique())>1}
    if len(colocated) == 0:
        return pd.DataFrame()
    # Add back date to each df
    for key, df in colocated.items():
        df[day_header] = [key for _ in range(len(df))]
    if merge == True:
        # Combine DataFrames
        return pd.concat(colocated.values(), axis=0)
    else:
        return colocated


def cluster_coords(df, lat_header, lon_header, precision):
    df = df.copy()
    # Make lat,lon hash column
    df['hash'] = calculate_geohashes(df[lat_header], df[lon_header], precision)
    # Make dict with hash:cluster, clusters need more than 1 point to count as a cluster
    return {key:cluster_df for key, cluster_df in df.groupby('hash') if len(cluster_df) > 1}


def colocation_clusters(clusters, id_header):
    """Return only clusters with more than one id"""
    return {key:df for key, df in clusters.items() if len(df[id_header].unique()) > 1}


def colocation_cluster_coords(df, lat_header, lon_header, id_header, precision):
    df = df.copy()
    # Make lat,lon hash column
    df['hash'] = calculate_geohashes(df[lat_header], df[lon_header], precision)
    # Make dict with hash:colocation cluster, clusters need more than 1 id to be a colocation cluster
    return {key:cluster_df for key, cluster_df in df.groupby('hash') if len(cluster_df[id_header].unique()) > 1}


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="hedron",
    version="0.0.5",
    description="A python package project for doing analysis on coordinates and clustering them.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/eddiethedean/hedron",
    author="Odos Matthews",
    author_email="odosmatthews@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["hedron"],
    include_package_data=True,
    install_requires=['pandas==1.2.4', 'PyGeodesy==21.6.9', 'py-staticmaps==0.4.0', 'range-key-dict==1.1.0'],
    entry_points={
        "console_scripts": [
            "hedron = hedron.cli:main",
        ]
    },
)